#include "Hooks_CreatedObject.h"
#include "Utilities.h"
#include "GameAPI.h"
#include "GameForms.h"

static FakeModInfo* g_FakeModInfo = NULL;

FakeModInfo::FakeModInfo(NVSESerializationInterface* _intfc)
{
	memset(this, 0, sizeof(FakeModInfo));
	intfc = _intfc;
	modIndex = 0xFF;

	// refModInfo is normally used to fix up formIDs as saved in GECK when referencing forms from other mods
	// for created objects, formIDs represent run-time formIDs, must be fixed up if load order changes
	DataHandler* dataHand = DataHandler::Get();
	numRefMods = dataHand->modList.loadedModCount;
	// TODO: here we need to build an array of mods based on previous load order

	// name, filepath, fileoffset, dataoffset
}

FakeModInfo::~FakeModInfo()
{
	//
}

FakeModInfo* FakeModInfo::Get() {
	if (!g_FakeModInfo)
		g_FakeModInfo = new FakeModInfo(&g_NVSESerializationInterface);
	return g_FakeModInfo;
}

UInt32 FakeModInfo::SetStaticFieldsAndGetFormTypeEnum(UInt32 aChunkType) {
	if (*s_ModInfo_CurrentChunkTypeCode == aChunkType)
		return *s_ModInfo_CurrentFormTypeEnum;
	for (UInt32 i = 0; i < 0x79; i++) {
		if (s_ModInfo_ChunkAndFormTypes[i].chunkType == aChunkType) {
			*s_ModInfo_CurrentChunkTypeCode = aChunkType;
			*s_ModInfo_CurrentFormTypeEnum = i;
			return i;
		}
	}
	return 0;
};

void SaveCreatedObjects(NVSESerializationInterface* fose)
{
	fose->OpenRecord('PERM', 0);
	UInt32 s = s_permanentClonedForms.size();
	fose->WriteRecordData(&s, sizeof(UInt32));
	for (UInt32 index = 0; index < s; index++) {
		PermanentClonedForm p = s_permanentClonedForms[index];
		fose->WriteRecordData(&p, sizeof(PermanentClonedForm));
	}

	TESSaveLoadGame* game = TESSaveLoadGame::Get();
	for (TESSaveLoadGame::CreatedObject* crobj = &game->createdObjectList; crobj; crobj = crobj->next)
	{
		if ((crobj->refID & 0xFF000000) == 0xFF000000)
		{
			TESForm* form = LookupFormByID(crobj->refID);
			if (!form)
			{
				_MESSAGE("SAVE: Unkown Object %08x found in created base object list", crobj->refID);
				continue;
			}
			else
			{

DumpClass(form, 0x80);
				form->SaveForm();
DumpClass(*g_CreatedObjectData, *g_CreatedObjectSize / 4);
				fose->OpenRecord('CROB', 0);
				fose->WriteRecordData(*g_CreatedObjectData, *g_CreatedObjectSize);
			}
		}
		else
			_MESSAGE("Save: Non-created object or garbage refID %08x found in created base object list.", crobj->refID);
	}
}

void fakeModInfo_GetNextChunk(void) {
	FakeModInfo* fakeModInfo = FakeModInfo::Get();
	ChunkHeader* current = (ChunkHeader*)((UInt32)(fakeModInfo->dataBuf)+fakeModInfo->dataOffset);
	while (current && ('XXXX' == current->type)) {
		// Skip this SubRecord
		fakeModInfo->dataOffset += current->size;
		if (fakeModInfo->dataOffset >= fakeModInfo->formInfo.dataSize)
			current = NULL;
		else
			current = (ChunkHeader*)((UInt32)(fakeModInfo->dataBuf)+fakeModInfo->dataOffset);
	};
	if (current) {
		fakeModInfo->subRecord.type = current->type;
		fakeModInfo->subRecord.size = current->size;
	}
};

void __declspec( naked ) Hook_fakeModInfo_GetNextChunk(void) {
	__asm {
		pusha
		call fakeModInfo_GetNextChunk
		popa
	}
};

static std::string aDirectory;
static std::string aName;

void SplitSavePath(const char* SavePath)
{

	std::string aTmp = std::string(SavePath);
	// truncate at last slash
	std::string::size_type	lastSlash = aTmp.rfind('\\');
	if(lastSlash != std::string::npos)	// if we don't find a slash something is VERY WRONG
	{
		aDirectory = aTmp.substr(0, lastSlash +1);
		aName = aTmp.substr(lastSlash + 1);
	}
	else
		{
			aDirectory.clear();
			aName.clear();
		}
}

bool LoadPermanentClonedForms(NVSESerializationInterface* fose) {
	s_permanentClonedForms.clear();
	UInt32 s;
	fose->ReadRecordData(&s, sizeof(UInt32));
	s_permanentClonedForms.resize(s);
	for (UInt32 index = 0; index < s; index++) {
		PermanentClonedForm p;
		fose->ReadRecordData(&p, sizeof(PermanentClonedForm));
		s_permanentClonedForms[index] = p;
	}
	return true;
}

bool LoadCreatedObject(NVSESerializationInterface* fose)
{
	//stub
	FakeModInfo* fakeModInfo = FakeModInfo::Get();
	DataHandler* dataHandler =DataHandler::Get();
	SplitSavePath(fakeModInfo->intfc->GetSavePath());
	strcpy_s(fakeModInfo->name, aName.c_str());
	strcpy_s(fakeModInfo->filepath, aDirectory.c_str());

	fose->PeekRecordData(&(fakeModInfo->formInfo), sizeof(fakeModInfo->formInfo));
	if (IsBigEndian()) {
		MACRO_SWAP32(fakeModInfo->formInfo.recordType);
		MACRO_SWAP32(fakeModInfo->formInfo.dataSize);
		MACRO_SWAP32(fakeModInfo->formInfo.formFlags);
		MACRO_SWAP32(fakeModInfo->formInfo.formID);
		MACRO_SWAP32(fakeModInfo->formInfo.unk10);
		MACRO_SWAP16(fakeModInfo->formInfo.formVersion);
		MACRO_SWAP16(fakeModInfo->formInfo.unk16);
	}
	fakeModInfo->SetStaticFieldsAndGetFormTypeEnum(fakeModInfo->formInfo.recordType);

	bool result = false;
	void* formData = FormHeap_Allocate(fakeModInfo->formInfo.dataSize);
	if (formData)
		__try {
			if (fose->ReadRecordData(formData, fakeModInfo->formInfo.dataSize) == fakeModInfo->formInfo.dataSize) 
				__try {
					fakeModInfo->dataBuf = (UInt8*)formData + sizeof(fakeModInfo->formInfo);
					fakeModInfo->dataBufSize = fakeModInfo->formInfo.dataSize /* - sizeof(fakeModInfo->formInfo) */;
					fakeModInfo->dataOffset = 0;
					fakeModInfo->fileOffset = 0;
					void** formDataAddr = &(fakeModInfo->dataBuf);
					g_CreatedObjectSize = &(fakeModInfo->dataBufSize);
					g_CreatedObjectData = (UInt8**)formDataAddr;

					TESForm* orgForm = NULL;
					TESForm* form = CreateFormInstance(*s_ModInfo_CurrentFormTypeEnum);
					if (!form) {
						_MESSAGE("LOAD: Unkown Object type %08x found in created objects cosave", s_ModInfo_CurrentFormTypeEnum);
					}
					else {
						for (UInt32 index = 0; index < s_permanentClonedForms.size(); index++)
							if (s_permanentClonedForms[index].cloneRefID = fakeModInfo->formInfo.formID) {
								orgForm = LookupFormByID(s_permanentClonedForms[index].orgRefID);
								if (orgForm)
									break;
							}

						if (orgForm) {
							UInt32 bWasCompressed = fakeModInfo->formInfo.formFlags & TESForm::kFormFlags_Compressed;
							UInt8 bSaveUnk61A = dataHandler->unk61A & 1;
							bool formReady = false;
							__try {
								dataHandler->unk61A |= 1;
								fakeModInfo->formInfo.formFlags |= TESForm::kFormFlags_Compressed;	// avoids calling TESFile.read in GetNextChunk...
								//form->CopyFrom(orgForm);
								form->LoadForm(fakeModInfo);
								form->InitItem();
								formReady = true;
							} __finally {
								if (!bWasCompressed)
									fakeModInfo->formInfo.formFlags &= !TESForm::kFormFlags_Compressed;	// restores
								if (!bSaveUnk61A) 
									dataHandler->unk61A &= 0xFFFFFFFE;
							}
							if (formReady)
								orgForm->DoAddForm(form, true, false);
						}
					}
			} __finally {
					fakeModInfo->dataBuf = NULL;
					fakeModInfo->dataOffset = 0;
					g_CreatedObjectSize = NULL;
					g_CreatedObjectData = NULL;
					result = true;
			}
		} __finally {
			FormHeap_Free(formData);
		}

	return result;
}
